package v5.name.panitz.wip;

import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;

import v5.name.panitz.wip.Tree.ConDef;
import v5.name.panitz.wip.Tree.FieldSelect;
import v5.name.panitz.wip.Tree.FunCall;
import v5.name.panitz.wip.Tree.FunDef;
import v5.name.panitz.wip.Tree.IntLiteral;
import v5.name.panitz.wip.Tree.OpExpr;
import v5.name.panitz.wip.Tree.Param;
import v5.name.panitz.wip.Tree.Program;
import v5.name.panitz.wip.Tree.Variable;

public class GenCode implements Visitor<Void> {
	Writer w;
	Map<String, Integer> vars;

	public GenCode() {
		try {
			w = new FileWriter("main.s");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	void finish() {
		try {
			w.write("  ret\n");
			w.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} 
		
	}
	
	@Override
	public Void visit(IntLiteral il) {
		// TODO Auto-generated method stub
		try {
			w.write("  movq $"+il.n+", %rax\n");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Void visit(OpExpr oe)  {
		try {
			oe.lhs.welcome(this);
			w.write("  pushq %rax\n");
			oe.rhs.welcome(this);
			w.write("  popq %rbx\n");
			switch (oe.op) {
			case mult: {
				w.write("  imulq %rbx, %rax\n"); 
				break;
			}
			default:
				w.write("  addq %rbx, %rax\n"); 
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Void visit(ConDef conDef) {
		try {
			w.append("\n  .globl "+conDef.name+"\n");
			w.append("  .type "+conDef.name+", @function\n\n");

			w.append(conDef.name+":\n");
			w.append("  pushq %rbp\n");
			w.append("  movq %rsp, %rbp\n");

			w.append("  movl	$"+8+", %edi\n");
			w.append("  call	malloc@PLT\n");

			var i = conDef.args.size()+1;
			var j = 0;
			for (var a:conDef.args) {
				//vars.put(a.name, (i--)*8);
				w.append("  movq	"+(i--)*8+"(%rbp), %rdx\n");
				w.append("  movq	%rdx, "+j+"(%rax)\n");
				j+=8;
			}
			
			w.append("  movq %rbp, %rsp\n");
			w.append("  popq %rbp\n");
			w.append("  ret\n\n");
			
		}catch (Exception e) {
			// TODO: handle exception
		}
		return null;
	}

	@Override
	public Void visit(FunDef funDef) {
		try {
			w.append("\n  .globl "+funDef.name+"\n");
			w.append("  .type "+funDef.name+", @function\n\n");
			
			w.append(funDef.name+":\n");
			w.append("  pushq %rbp\n");
			w.append("  movq %rsp, %rbp\n");

			vars = new HashMap<>();
			var i = funDef.args.size()+1;
			for (var a:funDef.args)vars.put(a.name, (i--)*8);
			
			funDef.body.welcome(this);
			
			w.append("  movq %rbp, %rsp\n");
			w.append("  popq %rbp\n");
			w.append("  ret\n\n");
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return null;
	}
	@Override
	public Void visit(Program program) {
		program.cons.forEach(con->con.welcome(this));
		program.funs.forEach(fun->fun.welcome(this));
		try {
			w.write(".globl	start\n");
			w.write(".type	start, @function\n");

			w.write("start:\n"); 
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		program.main.welcome(this);
		return null;
	}
	@Override
	public Void visit(Variable variable) {
		try {
			w.append("  movq "+vars.get(variable.name)+"(%rbp), %rax\n");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}
	@Override
	public Void visit(FunCall funCall) {
		funCall.params.forEach((param)->{
			param.welcome(this);
			try {
				w.append("  pushq %rax\n");
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		});
		try {
			w.append("  call "+funCall.name+"\n");
			w.append("  addq $"+(8*funCall.params.size())+", %rsp\n");

		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}
	@Override
	public Void visit(Param param) {
		// TODO Auto-generated method stub
		return null;
	}
	@Override
	public Void visit(FieldSelect fieldSelect) {
		fieldSelect.o.welcome(this);
		
		try {
			w.append("  movq	8(%rax), %rax\n");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return null;
	}

}
